# Copyright 2011-2014 Kwant authors.
#
# This file is part of Kwant.  It is subject to the license terms in the file
# LICENSE.rst found in the top-level directory of this distribution and at
# http://kwant-project.org/license.  A list of Kwant authors can be found in
# the file AUTHORS.rst at the top-level directory of this distribution and at
# http://kwant-project.org/authors.


from math import cos, sin
import numpy as np
from pytest import raises
from numpy.testing import assert_almost_equal
import kwant
from kwant._common import ensure_rng

n = 5
chain = kwant.lattice.chain()
sq = square = kwant.lattice.square()


class LeadWithOnlySelfEnergy:
    def __init__(self, lead):
        self.lead = lead
        self.parameters = frozenset()

    def selfenergy(self, energy, args=(), *, params=None):
        assert args == ()
        assert params == None
        return self.lead.selfenergy(energy)


def assert_modes_equal(modes1, modes2):
    assert_almost_equal(modes1.velocities, modes2.velocities)
    assert_almost_equal(modes1.momenta, modes2.momenta)
    vecs1, vecs2 = modes1.wave_functions, modes2.wave_functions
    assert_almost_equal(np.abs(np.sum(vecs1/vecs2, axis=0)),
                        vecs1.shape[0])


# Test output sanity: that an error is raised if no output is requested,
# and that solving for a subblock of a scattering matrix is the same as taking
# a subblock of the full scattering matrix.
def test_output(smatrix):
    rng = ensure_rng(3)
    system = kwant.Builder()
    left_lead = kwant.Builder(kwant.TranslationalSymmetry((-1,)))
    right_lead = kwant.Builder(kwant.TranslationalSymmetry((1,)))
    for b, site in [(system, chain(0)), (system, chain(1)),
                    (left_lead, chain(0)), (right_lead, chain(0))]:
        h = rng.random_sample((n, n)) + 1j * rng.random_sample((n, n))
        h += h.conjugate().transpose()
        b[site] = h
    for b, hopp in [(system, (chain(0), chain(1))),
                    (left_lead, (chain(0), chain(1))),
                    (right_lead, (chain(0), chain(1)))]:
        b[hopp] = (10 * rng.random_sample((n, n)) +
                   1j * rng.random_sample((n, n)))
    system.attach_lead(left_lead)
    system.attach_lead(right_lead)
    fsyst = system.finalized()

    result1 = smatrix(fsyst)
    s, modes1 = result1.data, result1.lead_info
    assert s.shape == 2 * (sum(len(i.momenta) for i in modes1) // 2,)
    s1 = result1.submatrix(1, 0)
    result2 = smatrix(fsyst, 0, (), [1], [0])
    s2, modes2 = result2.data, result2.lead_info
    assert s2.shape == (len(modes2[1].momenta) // 2,
                        len(modes2[0].momenta) // 2)
    assert_almost_equal(abs(s1), abs(s2))
    assert_almost_equal(np.dot(s.T.conj(), s),
                        np.identity(s.shape[0]))
    raises(ValueError, smatrix, fsyst, out_leads=[])
    modes = smatrix(fsyst).lead_info
    h = fsyst.leads[0].cell_hamiltonian()
    t = fsyst.leads[0].inter_cell_hopping()
    modes1 = kwant.physics.modes(h, t)[0]
    h = fsyst.leads[1].cell_hamiltonian()
    t = fsyst.leads[1].inter_cell_hopping()
    modes2 = kwant.physics.modes(h, t)[0]
    assert_modes_equal(modes1, modes[0])
    assert_modes_equal(modes2, modes[1])


# Test that a system with one lead has unitary scattering matrix.
def test_one_lead(smatrix):
    rng = ensure_rng(3)
    system = kwant.Builder()
    lead = kwant.Builder(kwant.TranslationalSymmetry((-1,)))
    for b, site in [(system, chain(0)), (system, chain(1)),
                    (system, chain(2)), (lead, chain(0))]:
        h = rng.random_sample((n, n)) + 1j * rng.random_sample((n, n))
        h += h.conjugate().transpose()
        b[site] = h
    for b, hopp in [(system, (chain(0), chain(1))),
                    (system, (chain(1), chain(2))),
                    (lead, (chain(0), chain(1)))]:
        b[hopp] = (10 * rng.random_sample((n, n)) +
                   1j * rng.random_sample((n, n)))
    system.attach_lead(lead)
    fsyst = system.finalized()

    for syst in (fsyst, fsyst.precalculate(), fsyst.precalculate(what='all')):
        s = smatrix(syst).data
        assert_almost_equal(np.dot(s.conjugate().transpose(), s),
                            np.identity(s.shape[0]))

    raises(ValueError, smatrix, fsyst.precalculate(what='selfenergy'))

# Test that a system with one lead with no propagating modes has a
# 0x0 S-matrix.
def test_smatrix_shape(smatrix):
    system = kwant.Builder()
    lead0 = kwant.Builder(kwant.TranslationalSymmetry((-1,)))
    lead1 = kwant.Builder(kwant.TranslationalSymmetry((1,)))
    for b, site in [(system, chain(0)), (system, chain(1)),
                    (system, chain(2))]:
        b[site] = 2
    lead0[chain(0)] = lambda site: lead0_val
    lead1[chain(0)] = lambda site: lead1_val

    for b, hopp in [(system, (chain(0), chain(1))),
                    (system, (chain(1), chain(2))),
                    (lead0, (chain(0), chain(1))),
                    (lead1, (chain(0), chain(1)))]:
        b[hopp] = -1
    system.attach_lead(lead0)
    system.attach_lead(lead1)
    fsyst = system.finalized()

    lead0_val = 4
    lead1_val = 4
    s = smatrix(fsyst, 1.0, (), [1], [0]).data
    assert s.shape == (0, 0)

    lead0_val = 2
    lead1_val = 2
    s = smatrix(fsyst, 1.0, (), [1], [0]).data
    assert s.shape == (1, 1)

    lead0_val = 4
    lead1_val = 2
    s = smatrix(fsyst, 1.0, (), [1], [0]).data
    assert s.shape == (1, 0)

    lead0_val = 2
    lead1_val = 4
    s = smatrix(fsyst, 1.0, (), [1], [0]).data
    assert s.shape == (0, 1)


# Test that a translationally invariant system with two leads has only
# transmission and that transmission does not mix modes.
def test_two_equal_leads(smatrix):
    def check_fsyst(fsyst):
        sol = smatrix(fsyst)
        s, leads = sol.data, sol.lead_info
        assert_almost_equal(np.dot(s.conjugate().transpose(), s),
                            np.identity(s.shape[0]))
        n_modes = len(leads[0].momenta) // 2
        assert len(leads[1].momenta) // 2 == n_modes
        assert_almost_equal(s[: n_modes, : n_modes], 0)
        t_elements = np.sort(abs(np.asarray(s[n_modes :, : n_modes])),
                             axis=None)
        t_el_should_be = n_modes * (n_modes - 1) * [0] + n_modes * [1]
        assert_almost_equal(t_elements, t_el_should_be)
        assert_almost_equal(sol.transmission(1,0), n_modes)
    rng = ensure_rng(11)
    system = kwant.Builder()
    lead = kwant.Builder(kwant.TranslationalSymmetry((1,)))
    h = rng.random_sample((n, n)) + 1j * rng.random_sample((n, n))
    h += h.conjugate().transpose()
    h *= 0.8
    t = 4 * rng.random_sample((n, n)) + 4j * rng.random_sample((n, n))
    lead[chain(0)] = system[chain(0)] = h
    lead[chain(0), chain(1)] = t
    system.attach_lead(lead)
    system.attach_lead(lead.reversed())
    fsyst = system.finalized()
    for syst in (fsyst, fsyst.precalculate(), fsyst.precalculate(what='all')):
        check_fsyst(syst)
    raises(ValueError, check_fsyst, fsyst.precalculate(what='selfenergy'))

    # Test the same, but with a larger scattering region.
    system = kwant.Builder()
    system[[chain(0), chain(1)]] = h
    system[chain(0), chain(1)] = t
    system.attach_lead(lead)
    system.attach_lead(lead.reversed())
    fsyst = system.finalized()
    for syst in (fsyst, fsyst.precalculate(), fsyst.precalculate(what='all')):
        check_fsyst(syst)
    raises(ValueError, check_fsyst, fsyst.precalculate(what='selfenergy'))


# Test a more complicated graph with non-singular hopping.
def test_graph_system(smatrix):
    rng = ensure_rng(11)
    system = kwant.Builder()
    lead = kwant.Builder(kwant.TranslationalSymmetry((-1, 0)))
    h = rng.random_sample((n, n)) + 1j * rng.random_sample((n, n))
    h += h.conjugate().transpose()
    h *= 0.8
    t = 4 * rng.random_sample((n, n)) + 4j * rng.random_sample((n, n))
    t1 = 4 * rng.random_sample((n, n)) + 4j * rng.random_sample((n, n))
    lead[sq(0, 0)] = system[[sq(0, 0), sq(1, 0)]] = h
    lead[sq(0, 1)] = system[[sq(0, 1), sq(1, 1)]] = 4 * h
    for builder in [system, lead]:
        builder[sq(0, 0), sq(1, 0)] = t
        builder[sq(0, 1), sq(1, 0)] = t1
        builder[sq(0, 1), sq(1, 1)] = 1.1j * t1
    system.attach_lead(lead)
    system.attach_lead(lead.reversed())
    fsyst = system.finalized()

    result = smatrix(fsyst)
    s, leads = result.data, result.lead_info
    assert_almost_equal(np.dot(s.conjugate().transpose(), s),
                        np.identity(s.shape[0]))
    n_modes = len(leads[0].momenta) // 2
    assert len(leads[1].momenta) // 2 == n_modes
    assert_almost_equal(s[: n_modes, : n_modes], 0)
    t_elements = np.sort(abs(np.asarray(s[n_modes:, :n_modes])),
                         axis=None)
    t_el_should_be = n_modes * (n_modes - 1) * [0] + n_modes * [1]
    assert_almost_equal(t_elements, t_el_should_be)


# Test a system with singular hopping.
def test_singular_graph_system(smatrix):
    rng = ensure_rng(11)

    system = kwant.Builder()
    lead = kwant.Builder(kwant.TranslationalSymmetry((-1, 0)))
    h = rng.random_sample((n, n)) + 1j * rng.random_sample((n, n))
    h += h.conjugate().transpose()
    h *= 0.8
    t = 4 * rng.random_sample((n, n)) + 4j * rng.random_sample((n, n))
    t1 = 4 * rng.random_sample((n, n)) + 4j * rng.random_sample((n, n))
    lead[sq(0, 0)] = system[[sq(0, 0), sq(1, 0)]] = h
    lead[sq(0, 1)] = system[[sq(0, 1), sq(1, 1)]] = 4 * h
    for builder in [system, lead]:
        builder[sq(0, 0), sq(1, 0)] = t
        builder[sq(0, 1), sq(1, 0)] = t1
    system.attach_lead(lead)
    system.attach_lead(lead.reversed())
    fsyst = system.finalized()

    result = smatrix(fsyst)
    s, leads = result.data, result.lead_info
    assert_almost_equal(np.dot(s.conjugate().transpose(), s),
                        np.identity(s.shape[0]))
    n_modes = len(leads[0].momenta) // 2
    assert len(leads[1].momenta) // 2 == n_modes
    assert_almost_equal(s[: n_modes, : n_modes], 0)
    t_elements = np.sort(abs(np.asarray(s[n_modes :, : n_modes])),
                         axis=None)
    t_el_should_be = n_modes * (n_modes - 1) * [0] + n_modes * [1]
    assert_almost_equal(t_elements, t_el_should_be)


# This test features inside the cell Hamiltonian a hopping matrix with more
# zero eigenvalues than the lead hopping matrix. Older version of the
# sparse solver failed here.
def test_tricky_singular_hopping(smatrix):
    system = kwant.Builder()
    lead = kwant.Builder(kwant.TranslationalSymmetry((4, 0)))

    interface = []
    for i in range(n):
        site = sq(-1, i)
        interface.append(site)
        system[site] = 0
        for j in range(4):
            lead[sq(j, i)] = 0
    for i in range(n-1):
        system[sq(-1, i), sq(-1, i+1)] = -1
        for j in range(4):
            lead[sq(j, i), sq(j, i+1)] = -1
    for i in range(n):
        for j in range(4):
            lead[sq(j, i), sq(j+1, i)] = -1
    del lead[sq(1, 0), sq(2, 0)]

    system.leads.append(kwant.builder.BuilderLead(lead, interface))
    fsyst = system.finalized()

    s = smatrix(fsyst, -1.3).data
    assert_almost_equal(np.dot(s.conjugate().transpose(), s),
                        np.identity(s.shape[0]))


# Test the consistency of transmission and conductance_matrix for a four-lead
# system without time-reversal symmetry.
def test_many_leads(*factories):
    E=2.1
    B=0.01

    def phase(a, b):
        ap = a.pos
        bp = b.pos
        phase = -B * (0.5 * (ap[1] + bp[1]) * (bp[0] - ap[0]))
        return -complex(cos(phase), sin(phase))

    # Build a square system with four leads and a hole in the center.
    syst = kwant.Builder()
    syst[(sq(x, y) for x in range(-4, 4) for y in range(-4, 4)
         if x**2 + y**2 >= 2)] = 3
    syst[sq.neighbors()] = phase
    for r in [range(-4, -1), range(4)]:
        lead = kwant.Builder(kwant.TranslationalSymmetry([-1, 0]))
        lead[(sq(0, y) for y in r)] = 4
        lead[sq.neighbors()] = phase
        syst.attach_lead(lead)
        syst.attach_lead(lead.reversed())
    syst = syst.finalized()

    r4 = list(range(4))
    br = factories[0](syst, E, out_leads=r4, in_leads=r4)
    trans = np.array([[br._transmission(i, j) for j in r4] for i in r4])
    cmat = br.conductance_matrix()
    assert_almost_equal(cmat.sum(axis=0), [0] * 4)
    assert_almost_equal(cmat.sum(axis=1), [0] * 4)
    for i in r4:
        for j in r4:
            assert_almost_equal(
                (br.num_propagating(i) if i == j else 0) - cmat[i, j],
                trans[i, j])

    for out_leads, in_leads in [(r4, r4), ((1, 2, 3), r4), (r4, (0, 2, 3)),
                                ((1, 3), (1, 2, 3)), ((0, 2), (0, 1, 2)),
                                ((0, 1,), (1, 2)), ((3,), (3,))]:
        for f in factories:
            br = f(syst, E, out_leads=out_leads, in_leads=in_leads)
            if len(out_leads) == 3:
                out_leads = r4
            if len(in_leads) == 3:
                in_leads = r4
            for i in r4:
                for j in r4:
                    if i in out_leads and j in in_leads:
                        assert_almost_equal(br.transmission(i, j), trans[i, j])
                    else:
                        raises(ValueError, br.transmission, i, j)
            if len(out_leads) == len(in_leads) == 4:
                assert_almost_equal(br.conductance_matrix(), cmat)


# Test equivalence between self-energy and scattering matrix representations.
# Also check that transmission works.
def test_selfenergy(greens_function, smatrix):
    rng = ensure_rng(4)
    system = kwant.Builder()
    left_lead = kwant.Builder(kwant.TranslationalSymmetry((-1,)))
    right_lead = kwant.Builder(kwant.TranslationalSymmetry((1,)))
    for b, site in [(system, chain(0)), (system, chain(1)),
                 (left_lead, chain(0)), (right_lead, chain(0))]:
        h = rng.random_sample((n, n)) + 1j * rng.random_sample((n, n))
        h += h.conjugate().transpose()
        b[site] = h
    for b, hopp in [(system, (chain(0), chain(1))),
                    (left_lead, (chain(0), chain(1))),
                    (right_lead, (chain(0), chain(1)))]:
        b[hopp] = (10 * rng.random_sample((n, n)) +
                   1j * rng.random_sample((n, n)))
    system.attach_lead(left_lead)
    system.attach_lead(right_lead)
    fsyst = system.finalized()

    t = smatrix(fsyst, 0, (), [1], [0]).data
    eig_should_be = np.linalg.eigvals(t * t.conjugate().transpose())
    n_eig = len(eig_should_be)

    def check_fsyst(fsyst):
        sol = greens_function(fsyst, 0, (), [1], [0])
        ttdagnew = sol._a_ttdagger_a_inv(1, 0)
        eig_are = np.linalg.eigvals(ttdagnew)
        t_should_be = np.sum(eig_are)
        assert_almost_equal(eig_are.imag, 0)
        assert_almost_equal(np.sort(eig_are.real)[-n_eig:],
                            np.sort(eig_should_be.real))
        assert_almost_equal(t_should_be, sol.transmission(1, 0))

    fsyst.leads[1] = LeadWithOnlySelfEnergy(fsyst.leads[1])
    check_fsyst(fsyst)

    fsyst.leads[0] = LeadWithOnlySelfEnergy(fsyst.leads[0])
    check_fsyst(fsyst)

    fsyst = system.finalized()
    for syst in (fsyst, fsyst.precalculate(what='selfenergy'),
                fsyst.precalculate(what='all')):
        check_fsyst(syst)
    raises(ValueError, check_fsyst, fsyst.precalculate(what='modes'))


def test_selfenergy_reflection(greens_function, smatrix):
    rng = ensure_rng(4)
    system = kwant.Builder()
    left_lead = kwant.Builder(kwant.TranslationalSymmetry((-1,)))
    for b, site in [(system, chain(0)), (system, chain(1)),
                 (left_lead, chain(0))]:
        h = rng.random_sample((n, n)) + 1j * rng.random_sample((n, n))
        h += h.conjugate().transpose()
        b[site] = h
    for b, hopp in [(system, (chain(0), chain(1))),
                    (left_lead, (chain(0), chain(1)))]:
        b[hopp] = (10 * rng.random_sample((n, n)) +
                   1j * rng.random_sample((n, n)))
    system.attach_lead(left_lead)
    fsyst = system.finalized()

    t = smatrix(fsyst, 0, (), [0], [0])

    fsyst.leads[0] = LeadWithOnlySelfEnergy(fsyst.leads[0])
    sol = greens_function(fsyst, 0, (), [0], [0])
    assert_almost_equal(sol.transmission(0,0), t.transmission(0,0))

    fsyst = system.finalized()
    for syst in (fsyst.precalculate(what='selfenergy'),
                fsyst.precalculate(what='all')):
        sol = greens_function(syst, 0, (), [0], [0])
        assert_almost_equal(sol.transmission(0,0), t.transmission(0,0))
    raises(ValueError, greens_function, fsyst.precalculate(what='modes'),
                  0, (), [0], [0])


def test_very_singular_leads(smatrix):
    syst = kwant.Builder()
    chain = kwant.lattice.chain()
    left_lead = kwant.Builder(kwant.TranslationalSymmetry((-1,)))
    right_lead = kwant.Builder(kwant.TranslationalSymmetry((1,)))
    syst[chain(0)] = left_lead[chain(0)] = right_lead[chain(0)] = np.identity(2)
    left_lead[chain(0), chain(1)] = np.zeros((2, 2))
    right_lead[chain(0), chain(1)] = np.identity(2)
    syst.attach_lead(left_lead)
    syst.attach_lead(right_lead)
    fsyst = syst.finalized()
    leads = smatrix(fsyst).lead_info
    assert [len(i.momenta) for i in leads] == [0, 4]


def test_ldos(ldos):
    syst = kwant.Builder()
    chain = kwant.lattice.chain()
    lead = kwant.Builder(kwant.TranslationalSymmetry(chain.vec((1,))))
    syst[chain(0)] = syst[chain(1)] = lead[chain(0)] = 0
    syst[chain(0), chain(1)] = lead[chain(0), chain(1)] = 1
    syst.attach_lead(lead)
    syst.attach_lead(lead.reversed())
    fsyst = syst.finalized()

    for finsyst in (fsyst, fsyst.precalculate(what='modes'),
                   fsyst.precalculate(what='all')):
        assert_almost_equal(ldos(finsyst, 0),
                            np.array([1, 1]) / (2 * np.pi))
    raises(ValueError, ldos, fsyst.precalculate(what='selfenergy'), 0)
    fsyst.leads[0] = LeadWithOnlySelfEnergy(fsyst.leads[0])
    raises(NotImplementedError, ldos, fsyst, 0)


def test_wavefunc_ldos_consistency(wave_function, ldos):
    L = 2
    W = 3

    rng = ensure_rng(31)
    syst = kwant.Builder()
    left_lead = kwant.Builder(kwant.TranslationalSymmetry((-1, 0)))
    top_lead = kwant.Builder(kwant.TranslationalSymmetry((1, 0)))
    for b, sites in [(syst, [square(x, y)
                               for x in range(L) for y in range(W)]),
                     (left_lead, [square(0, y) for y in range(W)]),
                     (top_lead, [square(x, 0) for x in range(L)])]:
        for site in sites:
            h = rng.random_sample((n, n)) + 1j * rng.random_sample((n, n))
            h += h.conjugate().transpose()
            b[site] = h
        for hopping_kind in square.neighbors():
            for hop in hopping_kind(b):
                b[hop] = (10 * rng.random_sample((n, n)) +
                          1j * rng.random_sample((n, n)))
    syst.attach_lead(left_lead)
    syst.attach_lead(top_lead)
    syst = syst.finalized()

    def check(syst):
        for energy in [0, 1000]:
            wf = wave_function(syst, energy)
            ldos2 = np.zeros(wf.num_orb, float)
            for lead in range(len(syst.leads)):
                temp = abs(wf(lead))
                temp **= 2
                ldos2 += temp.sum(axis=0)
            ldos2 *= (0.5 / np.pi)

            assert_almost_equal(ldos2, ldos(syst, energy))

    for fsyst in (syst, syst.precalculate(what='modes'),
                 syst.precalculate(what='all')):
        check(fsyst)
    raises(ValueError, check, syst.precalculate(what='selfenergy'))
    syst.leads[0] = LeadWithOnlySelfEnergy(syst.leads[0])
    raises(NotImplementedError, check, syst)


def test_arg_passing(wave_function, ldos, smatrix):

    def onsite(site, a, b):
        return site.pos[0] + site.pos[1] + a + b

    def hopping(site1, site2, a, b):
        return b - a

    W = 3
    L = 4

    syst = kwant.Builder()
    syst[(square(i, j) for i in range(L) for j in range(W))] = onsite
    syst[square.neighbors()] = hopping

    lead = kwant.Builder(kwant.TranslationalSymmetry((-1, 0)))
    lead[(square(0, j) for j in range(W))] = onsite
    lead[square.neighbors()] = hopping

    syst.attach_lead(lead)
    syst.attach_lead(lead.reversed())

    fsyst = syst.finalized()

    # compare results to those when we pass `args` only
    args = (1, 3)
    params = dict(a=1, b=3)
    np.testing.assert_array_equal(
        wave_function(fsyst, args=args)(0),
        wave_function(fsyst, params=params)(0))
    np.testing.assert_array_equal(
        ldos(fsyst, args=args),
        ldos(fsyst, params=params))
    np.testing.assert_array_equal(
        smatrix(fsyst, args=args).data,
        smatrix(fsyst, params=params).data)



# Copyright 2011-2013 Kwant authors.
#
# This file is part of Kwant.  It is subject to the license terms in the file
# LICENSE.rst found in the top-level directory of this distribution and at
# http://kwant-project.org/license.  A list of Kwant authors can be found in
# the file AUTHORS.rst at the top-level directory of this distribution and at
# http://kwant-project.org/authors.

import numpy as np

class _Random:
    def __init__(self):
        self._x = 13

    def _set_seed(self, seed):
        self._x = seed

    def _randf(self):
        # A very bad random number generator returning numbers between -1 and
        # +1.  Just for making some matrices, and being sure that they are the
        # same on any architecture.
        m = 2**16
        a = 11929
        c = 36491

        self._x = (a * self._x + c) % m

        return (float(self._x)/m-0.5)*2

    def _randi(self):
        # A very bad random number generator returning number between 0 and 20.
        # Just for making some matrices, and being sure that they are the same
        # on any architecture.
        m = 2**16
        a = 11929
        c = 36491

        self._x = (a * self._x + c) % m

        return self._x % 21

    def randmat(self, n, m, dtype):
        mat = np.empty((n, m), dtype = dtype)

        if issubclass(dtype, np.complexfloating):
            for i in range(n):
                for j in range(m):
                    mat[i,j] = self._randf() + 1j * self._randf()
        elif issubclass(dtype, np.floating):
            for i in range(n):
                for j in range(m):
                    mat[i,j] = self._randf()
        elif issubclass(dtype, np.integer):
            for i in range(n):
                for j in range(m):
                    mat[i,j] = self._randi()

        return mat

    def randvec(self, n, dtype):
        vec = np.empty(n, dtype = dtype)

        if issubclass(dtype, np.complexfloating):
            for i in range(n):
                vec[i] = self._randf() + 1j * self._randf()
        elif issubclass(dtype, np.floating):
            for i in range(n):
                vec[i] = self._randf()
        elif issubclass(dtype, np.integer):
            for i in range(n):
                vec[i] = self._randi()

        return vec

# Improved version of assert_arrays_almost_equal that uses the dtype to set the
# precision (The default precision of assert_arrays_almost_equal is sometimes
# too small for single-precision comparisions.)
def assert_array_almost_equal(dtype, a, b):
    if dtype == np.float32 or dtype == np.complex64:
        prec = 5
    else:
        prec = 10

    np.testing.assert_array_almost_equal(a, b, decimal=prec)



# Copyright 2011-2017 Kwant authors.
#
# This file is part of Kwant.  It is subject to the license terms in the file
# LICENSE.rst found in the top-level directory of this distribution and at
# http://kwant-project.org/license.  A list of Kwant authors can be found in
# the file AUTHORS.rst at the top-level directory of this distribution and at
# http://kwant-project.org/authors.

try:
    from .discretizer import discretize, discretize_symbolic, build_discretized
    from ._common import sympify, lambdify
    from ._common import momentum_operators, position_operators
except ImportError as error:
    msg = ("'kwant.continuum' is not available because one or more of its "
           "dependencies is not installed.")
    raise ImportError(msg) from error


__all__ = ['discretize', 'discretize_symbolic', 'build_discretized',
           'sympify', 'lambdify', 'momentum_operators', 'position_operators']



# Copyright 2011-2015 Kwant authors.
#
# This file is part of Kwant.  It is subject to the license terms in the file
# LICENSE.rst found in the top-level directory of this distribution and at
# http://kwant-project.org/license.  A list of Kwant authors can be found in
# the file AUTHORS.rst at the top-level directory of this distribution and at
# http://kwant-project.org/authors.

import sys
import numpy as np
import numbers
import inspect
import warnings
import importlib
import functools
from contextlib import contextmanager

__all__ = ['KwantDeprecationWarning', 'UserCodeError']


class KwantDeprecationWarning(Warning):
    """Class of warnings about a deprecated feature of Kwant.

    DeprecationWarning has been made invisible by default in Python 2.7 in order
    to not confuse non-developer users with warnings that are not relevant to
    them.  In the case of Kwant, by far most users are developers, so we feel
    that a KwantDeprecationWarning that is visible by default is useful.
    """
    pass


class UserCodeError(Exception):
    """Class for errors that occur in user-provided code.

    Usually users will define value functions that Kwant calls in order to
    evaluate the Hamiltonian.  If one of these function raises an exception
    then it is caught and this error is raised in its place. This makes it
    clear that the error is from the user's code (and not a bug in Kwant) and
    also makes it possible for any libraries that wrap Kwant to detect when a
    user's function causes an error.
    """
    pass


def deprecate_parameter(parameter_name, version=None, help=None,
                        stacklevel=3):
    """Trigger a deprecation warning if the wrapped function is called
       with the provided parameter."""

    message = ("The '{}' parameter has been deprecated since version {} -- {}"
               .format(parameter_name, version, help))

    def warn():
        warnings.warn(message, KwantDeprecationWarning,
                      stacklevel=stacklevel)

    def wrapper(f=None):

        # Instead of being used as a decorator, can be called with
        # no arguments to just raise the warning.
        if f is None:
            warn()
            return

        sig = inspect.signature(f)

        @functools.wraps(f)
        def inner(*args, **kwargs):
            # If the named argument is truthy
            if sig.bind(*args, **kwargs).arguments.get(parameter_name):
                warn()
            return f(*args, **kwargs)

        return inner

    return wrapper


# Deprecation for 'args' parameter; defined once to minimize boilerplate,
# as this parameter is present all over Kwant.
deprecate_args = deprecate_parameter('args', version=1.4, help=
    "Instead, provide named parameters as a dictionary via 'params'.")


def interleave(seq):
    """Return an iterator that yields pairs of elements from a sequence.

    If 'seq' has an odd number of elements then the last element is dropped.

    Examples
    --------
    >>> list(interleave(range(4)))
    [(0, 1), (2, 3)]
    >>> list(interleave(range(5))
    [(0, 1), (2, 3)]
    """
    # zip, when given the same iterator twice, turns a sequence into a
    # sequence of pairs.
    iseq = iter(seq)
    return zip(iseq, iseq)


def ensure_isinstance(obj, typ, msg=None):
    if isinstance(obj, typ):
        return
    if msg is None:
        msg = "Expecting an instance of {}.".format(typ.__name__)
    raise TypeError(msg)

def ensure_rng(rng=None):
    """Turn rng into a random number generator instance

    If rng is None, return the RandomState instance used by np.random.
    If rng is an integer, return a new RandomState instance seeded with rng.
    If rng is already a RandomState instance, return it.
    Otherwise raise ValueError.
    """
    if rng is None:
        return np.random.mtrand._rand
    if isinstance(rng, numbers.Integral):
        return np.random.RandomState(rng)
    if all(hasattr(rng, attr) for attr in ('random_sample', 'randn',
                                           'randint', 'choice')):
        return rng
    raise ValueError("Expecting a seed or an object that offers the "
                     "numpy.random.RandomState interface.")


@contextmanager
def reraise_warnings(level=3):
    with warnings.catch_warnings(record=True) as caught_warnings:
        yield
    for warning in caught_warnings:
        warnings.warn(warning.message, stacklevel=level)


def get_parameters(func):
    """Return the names of parameters of a function.

    It is made sure that the function can be called as func(*args) with
    'args' corresponding to the returned parameter names.

    Returns
    -------
    param_names : list
        Names of positional parameters that appear in the signature of 'func'.
    """
    def error(msg):
        fname = inspect.getsourcefile(func)
        try:
            line = inspect.getsourcelines(func)[1]
        except OSError:
            line = '<unknown line>'
        raise ValueError("{}:\nFile {}, line {}, in {}".format(
            msg, repr(fname), line, func.__name__))

    P = inspect.Parameter
    pars = inspect.signature(func).parameters # an *ordered mapping*
    names = []
    for k, v in pars.items():
        if v.kind in (P.POSITIONAL_ONLY, P.POSITIONAL_OR_KEYWORD):
            if v.default is P.empty:
                names.append(k)
            else:
                error("Arguments of value functions "
                      "must not have default values")
        elif v.kind is P.KEYWORD_ONLY:
            error("Keyword-only arguments are not allowed in value functions")
        elif v.kind in (P.VAR_POSITIONAL, P.VAR_KEYWORD):
            error("Value functions must not take *args or **kwargs")
    return tuple(names)


class lazy_import:
    def __init__(self, module, package='kwant', deprecation_warning=False):
        if module.startswith('.') and not package:
            raise ValueError('Cannot import a relative module without a package.')
        self.__module = module
        self.__package = package
        self.__deprecation_warning = deprecation_warning

    def __getattr__(self, name):
        if self.__deprecation_warning:
            msg = ("Accessing {0} without an explicit import is deprecated. "
                   "Instead, explicitly 'import {0}'."
                  ).format('.'.join((self.__package, self.__module)))
            warnings.warn(msg, KwantDeprecationWarning, stacklevel=2)
        relative_module = '.' + self.__module
        mod = importlib.import_module(relative_module, self.__package)
        # Replace this _LazyModuleProxy with an actual module
        package = sys.modules[self.__package]
        setattr(package, self.__module, mod)
        return getattr(mod, name)



# Copyright 2011-2013 Kwant authors.
#
# This file is part of Kwant.  It is subject to the license terms in the file
# LICENSE.rst found in the top-level directory of this distribution and at
# http://kwant-project.org/license.  A list of Kwant authors can be found in
# the file AUTHORS.rst at the top-level directory of this distribution and at
# http://kwant-project.org/authors.

import numpy as np

int_dtype = np.int32



# Copyright 2011-2016 Kwant authors.
#
# This file is part of Kwant.  It is subject to the license terms in the file
# LICENSE.rst found in the top-level directory of this distribution and at
# http://kwant-project.org/license.  A list of Kwant authors can be found in
# the file AUTHORS.rst at the top-level directory of this distribution and at
# http://kwant-project.org/authors.

cimport numpy as np
from cpython cimport array
from .defs cimport gint

cdef struct Edge:
    gint tail, head

cdef class Graph:
    cdef int allow_negative_nodes
    cdef Edge *edges
    cdef gint capacity, size, _num_nodes
    cdef gint num_pp_edges, num_pn_edges, num_np_edges

    cpdef reserve(self, gint capacity)
    cpdef gint add_edge(self, gint tail, gint head) except -1
    cdef _add_edges_ndarray_int64(self, np.ndarray[np.int64_t, ndim=2] edges)
    cdef _add_edges_ndarray_int32(self, np.ndarray[np.int32_t, ndim=2] edges)

cdef class gintArraySlice:
    cdef gint *data
    cdef gint size

cdef class CGraph:
    cdef readonly bint twoway, edge_nr_translation
    cdef readonly gint num_nodes, num_edges, num_px_edges, num_xp_edges
    cdef array.array _heads_idxs
    cdef gint *heads_idxs
    cdef array.array _heads
    cdef gint *heads
    cdef array.array _tails_idxs
    cdef gint *tails_idxs
    cdef array.array _tails
    cdef gint *tails
    cdef array.array _edge_ids
    cdef gint *edge_ids
    cdef array.array _edge_ids_by_edge_nr
    cdef gint *edge_ids_by_edge_nr
    cdef gint edge_nr_end

    cpdef gintArraySlice out_neighbors(self, gint node)


cdef class CGraph_malloc(CGraph):
    pass

cdef class EdgeIterator:
    cdef CGraph graph
    cdef gint edge_id, tail



# Copyright 2011-2017 Kwant authors.
#
# This file is part of Kwant.  It is subject to the license terms in the file
# LICENSE.rst found in the top-level directory of this distribution and at
# http://kwant-project.org/license.  A list of Kwant authors can be found in
# the file AUTHORS.rst at the top-level directory of this distribution and at
# http://kwant-project.org/authors.

import inspect
import warnings
from functools import wraps

import numpy as np
import pytest

import sympy

from ..discretizer import discretize
from ..discretizer import discretize_symbolic
from ..discretizer import build_discretized
from ..discretizer import  _wf

from ...lattice import Monatomic


def swallows_extra_kwargs(f):
    sig = inspect.signature(f)
    pars = sig.parameters
    if any(i.kind is inspect.Parameter.VAR_KEYWORD for i in pars.values()):
        return f

    names = {name for name, value in pars.items() if
             value.kind not in (inspect.Parameter.VAR_POSITIONAL,
                                inspect.Parameter.POSITIONAL_ONLY)}

    @wraps(f)
    def wrapped(*args, **kwargs):
        bound = sig.bind(*args, **{name: value for name, value
                                   in kwargs.items() if name in names})
        return f(*bound.args, **bound.kwargs)

    return wrapped


I = sympy.I
kx, ky, kz = sympy.symbols('k_x k_y k_z', commutative=False)
x, y, z = sympy.symbols('x y z', commutative=False)
ax, ay, az = sympy.symbols('a_x a_y a_z')
a = sympy.symbols('a')

wf = _wf
Psi = wf(x, y, z)
A, B = sympy.symbols('A B', commutative=False)
fA, fB = sympy.symbols('A B', cls=sympy.Function)

ns = {'A': A, 'B': B, 'a_x': ax, 'a_y': ay, 'az': az, 'x': x, 'y': y, 'z': z}


@pytest.mark.parametrize('commutative', [True, False])
def test_reading_coordinates(commutative):
    kx, ky, kz = sympy.symbols('k_x k_y k_z', commutative=commutative)

    test = {
        kx**2                         : ['x'],
        kx**2 + ky**2                 : ['x', 'y'],
        kx**2 + ky**2 + kz**2         : ['x', 'y', 'z'],
        ky**2 + kz**2                 : ['y', 'z'],
        kz**2                         : ['z'],
        kx * fA(x, y) * kx              : ['x'],
        kx**2 + kz * fB(y)             : ['x', 'z'],
    }
    for inp, out in test.items():
        ham, got = discretize_symbolic(inp)
        assert got == out


def test_reading_coordinates_matrix():
    test = [
        (sympy.Matrix([sympy.sympify('k_x**2')])      , ['x']),
        (sympy.Matrix([kx**2])                        , ['x']),
        (sympy.Matrix([kx**2 + ky**2])                , ['x', 'y']),
        (sympy.Matrix([kx**2 + ky**2 + kz**2])        , ['x', 'y', 'z']),
        (sympy.Matrix([ky**2 + kz**2])                , ['y', 'z']),
        (sympy.Matrix([kz**2])                        , ['z']),
        (sympy.Matrix([kx * fA(x, y) * kx])            , ['x']),
        (sympy.Matrix([kx**2 + kz * fB(y)])            , ['x', 'z']),
    ]
    for inp, out in test:
        ham, got = discretize_symbolic(inp)
        assert got == out


def test_reading_different_matrix_types():
    test = [
        (sympy.MutableMatrix([kx**2])                    , ['x']),
        (sympy.ImmutableMatrix([kx**2])                  , ['x']),
        (sympy.MutableDenseMatrix([kx**2])               , ['x']),
        (sympy.ImmutableDenseMatrix([kx**2])             , ['x']),
    ]
    for inp, out in test:
        ham, got = discretize_symbolic(inp)
        assert got == out,\
            "Should be: _split_factors({})=={}. Not {}".format(inp, out, got)


@pytest.mark.parametrize('commutative', [True, False])
def test_simple_derivations(commutative):
    kx, ky, kz = sympy.symbols('k_x k_y k_z', commutative=commutative)
    test = {
        kx**2                   : {(0,): 2/ax**2, (1,): -1/ax**2},
        kx**2 + 1               : {(0,): 1 + 2/ax**2, (1,): -1/ax**2},
        kx**2 + 2               : {(0,): 2 + 2/ax**2, (1,): -1/ax**2},
        kx**2 + 1.0             : {(0,): 1.0 + 2/ax**2, (1,): -1/ax**2},
        kx**2 + 2.0             : {(0,): 2.0 + 2/ax**2, (1,): -1/ax**2},
        kx**2 + ky**2           : {(0, 1): -1/ay**2, (0, 0): 2/ax**2 + 2/ay**2,
                                   (1, 0): -1/ax**2},
        kx**2 + ky**2 + kz**2   : {(1, 0, 0): -1/ax**2, (0, 0, 1): -1/az**2,
                                   (0, 0, 0): 2/ax**2 + 2/ay**2 + 2/az**2,
                                   (0, 1, 0): -1/ay**2},
        ky**2 + kz**2           : {(0, 1): -1/az**2, (0, 0): 2/ay**2 + 2/az**2,
                                   (1, 0): -1/ay**2},
        kz**2                   : {(0,): 2/az**2, (1,): -1/az**2},
    }
    non_commutative_test = {
        kx * fA(x, y) * kx       : {(1, ): -fA(ax/2 + x, y)/ax**2,
                                  (0, ): fA(-ax/2 + x, y)/ax**2 + fA(ax/2 + x, y)/ax**2},
        kx**2 + kz * fB(y)       : {(1, 0): -1/ax**2, (0, 1): -I*fB(y)/(2*az),
                                   (0, 0): 2/ax**2},
        kx * fA(x)               : {(0,): 0, (1,): -I*fA(ax + x)/(2*ax)},
        ky * fA(x)               : {(1,): -I*fA(x)/(2*ay), (0,): 0},
        kx * fA(x) * B           : {(0,): 0, (1,): -I*fA(ax + x)*B/(2*ax)},
        5 * kx                  : {(0,): 0, (1,): -5*I/(2*ax)},
        kx * (fA(x) + fB(x))      : {(0,): 0,
                                   (1,): -I*fA(ax + x)/(2*ax) - I*fB(ax + x)/(2*ax)},
    }

    if not commutative:
        test.update(non_commutative_test)

    for inp, out in test.items():
        got, _ = discretize_symbolic(inp)
        assert got == out

    for inp, out in test.items():
        got, _ = discretize_symbolic(str(inp), locals=ns)
        assert got == out


@pytest.mark.parametrize('e_to_subs, e, subs', [
    ('A * k_x', '(A + B) * k_x', {'A': 'A + B'}),
    ('k_x', 'k_x + k_y', {'k_x': 'k_x + k_y'}),
    ('k_x**2 + V', 'k_x**2 + V + V_0', {'V': 'V + V_0'}),
    ('k_x**2 + A + C', 'k_x**2 + B + 5', {'A': 'B + 5', 'C': 0}),
    ('x + y + z', '1 + 3 + 5', {'x': 1, 'y': 3, 'z': 5}),
])
def test_simple_derivations_with_subs(e_to_subs, e, subs):
    one = discretize_symbolic(e_to_subs, 'xyz', locals=subs)
    two = discretize_symbolic(e, 'xyz')
    assert one == two


def test_simple_derivations_matrix():
    test = {
        kx**2                   : {(0,): 2/ax**2, (1,): -1/ax**2},
        kx**2 + ky**2           : {(0, 1): -1/ay**2, (0, 0): 2/ax**2 + 2/ay**2,
                                   (1, 0): -1/ax**2},
        kx**2 + ky**2 + kz**2   : {(1, 0, 0): -1/ax**2, (0, 0, 1): -1/az**2,
                                   (0, 0, 0): 2/ax**2 + 2/ay**2 + 2/az**2,
                                   (0, 1, 0): -1/ay**2},
        ky**2 + kz**2           : {(0, 1): -1/az**2, (0, 0): 2/ay**2 + 2/az**2,
                                   (1, 0): -1/ay**2},
        kz**2                   : {(0,): 2/az**2, (1,): -1/az**2},

        kx * fA(x, y) * kx       : {(1, ): -fA(ax/2 + x, y)/ax**2,
                                  (0, ): fA(-ax/2 + x, y)/ax**2 + fA(ax/2 + x, y)/ax**2},
        kx**2 + kz * fB(y)       : {(1, 0): -1/ax**2, (0, 1): -I*fB(y)/(2*az),
                                   (0, 0): 2/ax**2},
        kx * fA(x)               : {(0,): 0, (1,): -I*fA(ax + x)/(2*ax)},
        ky * fA(x)               : {(1,): -I*fA(x)/(2*ay), (0,): 0},
        kx * fA(x) * B           : {(0,): 0, (1,): -I*fA(ax + x)*B/(2*ax)},
        5 * kx                  : {(0,): 0, (1,): -5*I/(2*ax)},
        kx * (fA(x) + fB(x))      : {(0,): 0,
                                   (1,): -I*fA(ax + x)/(2*ax) - I*fB(ax + x)/(2*ax)},
   }

    new_test = []
    for inp, out in test.items():
        new_out = {}
        for k, v in out.items():
            new_out[k] = sympy.Matrix([v])
        new_test.append((sympy.Matrix([inp]), new_out))

    for inp, out in new_test:
        got, _ = discretize_symbolic(inp)
        assert got == out

    for inp, out in new_test:
        got, _ = discretize_symbolic(str(inp), locals=ns)
        assert got == out

    for inp, out in new_test:
        got, _ = discretize_symbolic(str(inp).replace('Matrix', ''), locals=ns)
        assert got == out



def test_integer_float_input():
    test = {
        0: {(0, 0, 0): 0},
        1: {(0, 0, 0): 1},
        5: {(0, 0, 0): 5},
    }

    for inp, out in test.items():
        got, _ = discretize_symbolic(int(inp), 'xyz')
        assert got == out

        got, _ = discretize_symbolic(float(inp), 'xyz')
        assert got == out

    # let's test in matrix version too
    new_test = []
    for inp, out in test.items():
        new_out = {}
        for k, v in out.items():
            new_out[k] = sympy.Matrix([v])
        new_test.append((inp, new_out))

    for inp, out in new_test:
        got, _ = discretize_symbolic(sympy.Matrix([int(inp)]), 'xyz')
        assert got == out

        got, _ = discretize_symbolic(sympy.Matrix([float(inp)]), 'xyz')
        assert got == out


def test_different_discrete_coordinates():
    test = [
        (
            'xyz', {
                (1, 0, 0): -1/ax**2, (0, 0, 1): -1/az**2,
                (0, 0, 0): 2/ax**2 + 2/ay**2 + 2/az**2, (0, 1, 0): -1/ay**2
            }
        ),
        (
            'xy', {
                (0, 1): -1/ay**2,
                (1, 0): -1/ax**2,
                (0, 0): kz**2 + 2/ax**2 + 2/ay**2
            }
        ),
        (
            'xz', {
                (0, 1): -1/az**2,
                (1, 0): -1/ax**2,
                (0, 0): ky**2 + 2/ax**2 + 2/az**2
            }
        ),
        (
            'yz', {
                (0, 1): -1/az**2,
                (1, 0): -1/ay**2,
                (0, 0): kx**2 + 2/ay**2 + 2/az**2
            }
        ),
        (
            'x', {
                (0,): ky**2 + kz**2 + 2/ax**2, (1,): -1/ax**2
            }
        ),
        (
            'y', {
                (0,): kx**2 + kz**2 + 2/ay**2, (1,): -1/ay**2
            }
        ),
        (
            'z', {
                (0,): ky**2 + kx**2 + 2/az**2, (1,): -1/az**2
            }
        ),
    ]
    for inp, out in test:
        got, _ = discretize_symbolic(kx**2 + ky**2 + kz**2, inp)
        assert got == out

    # let's test in matrix version too
    new_test = []
    for inp, out in test:
        new_out = {}
        for k, v in out.items():
            new_out[k] = sympy.Matrix([v])
        new_test.append((inp, new_out))

    for inp, out in new_test:
        got, _ = discretize_symbolic(sympy.Matrix([kx**2 + ky**2 + kz**2]), inp)
        assert got == out


def test_non_expended_input():
    symbolic, coords = discretize_symbolic(kx * (kx + fA(x)))
    desired = {
        (0,): 2/ax**2,
        (1,): -I*fA(ax + x)/(2*ax) - 1/ax**2
    }
    assert symbolic == desired


def test_matrix_with_zeros():
    Matrix = sympy.Matrix
    symbolic, _ = discretize_symbolic("[[k_x*A(x)*k_x, 0], [0, k_x*A(x)*k_x]]")
    output = {
        (0,):  Matrix([[fA(-ax/2 + x)/ax**2 + fA(ax/2 + x)/ax**2, 0], [0, fA(-ax/2 + x)/ax**2 + fA(ax/2 + x)/ax**2]]),
        (1,):  Matrix([[-fA(ax/2 + x)/ax**2, 0], [0, -fA(ax/2 + x)/ax**2]]),
        }
    assert symbolic == output


def test_numeric_functions_basic_symbolic():
    for i in [0, 1, 3, 5]:
        builder = discretize(i, 'x')
        lat = next(iter(builder.sites()))[0]
        assert builder.lattice is lat
        assert builder[lat(0)] == i

        p = dict(t=i)

        tb = {(0,): sympy.sympify("2*t"), (1,): sympy.sympify('-t')}
        builder = build_discretized(tb, 'x', grid=1)
        lat = next(iter(builder.sites()))[0]
        assert 2*p['t'] == builder[lat(0)](None, **p)
        assert -p['t'] == builder[lat(1), lat(0)](None, None, **p)

        tb = {(0,): sympy.sympify("0"), (1,): sympy.sympify('-1j * t')}
        builder = build_discretized(tb, 'x', grid=1)
        lat = next(iter(builder.sites()))[0]
        assert -1j * p['t'] == builder[lat(0), lat(1)](None, None, **p)
        assert +1j * p['t'] == builder[lat(1), lat(0)](None, None, **p)


@pytest.mark.parametrize('commutative', [True, False])
def test_numeric_function_coords_from_site(commutative):
    tb = {(0,): sympy.symbols('x', commutative=commutative)}
    builder = build_discretized(tb, 'x')

    lat = next(iter(builder.sites()))[0]
    onsite = builder[lat(0)]
    assert (onsite(lat(0)) == 0 and onsite(lat(1)) == 1)


def test_numeric_functions_not_discrete_coords():
    builder = discretize('k_y + y', 'x')
    lat = next(iter(builder.sites()))[0]
    onsite = builder[lat(0)]

    assert onsite(None, k_y=2, y=1) == 2 + 1


@pytest.mark.parametrize('ham, val, params', [
    ("pi", np.pi, {}),
    ("A + pi", 1 + np.pi, {"A": 1}),
    ("A + B(pi)", 1 + np.pi, {"A": 1, "B": lambda x: x}),
    ("A + I", 1 + 1j, {"A": 1}),
    ("A + 1j", 1 + 1j, {"A": 1}),
    ("A + B(I)", 1 + 1j, {"A": 1, "B": lambda x: x}),
    ("A + B(1j)", 1 + 1j, {"A": 1, "B": lambda x: x}),
    ("exp(1j * pi)", np.exp(1j*np.pi), {"exp": np.exp}),
    (sympy.exp(sympy.sympify("1j * pi * A")), np.exp(1j*np.pi),
        {"exp": np.exp, "A": 1}),
])
def test_numeric_functions_advanced(ham, val, params):
    builder = discretize(ham, 'x')
    lat = next(iter(builder.sites()))[0]
    onsite = builder[lat(0)]
    try:
        assert np.allclose(onsite(None, **params), val)
    except TypeError:
        assert np.allclose(onsite, val)


def test_numeric_functions_basic_string():
    for i in [0, 1, 3, 5]:
        builder = discretize(i, 'x')
        lat = next(iter(builder.sites()))[0]
        assert builder[lat(0)] == i

        p = dict(t=i)

        tb = {(0,): "2*t", (1,): "-t"}
        builder = build_discretized(tb, 'x', grid=1)
        lat = next(iter(builder.sites()))[0]
        assert 2*p['t'] == builder[lat(0)](None, **p)
        assert -p['t'] == builder[lat(1), lat(0)](None, None, **p)

        tb = {(0,): "0", (1,): "-1j * t"}
        builder = build_discretized(tb, 'x', grid=1)
        lat = next(iter(builder.sites()))[0]
        assert -1j * p['t'] == builder[lat(0), lat(1)](None, None, **p)
        assert +1j * p['t'] == builder[lat(1), lat(0)](None, None, **p)

        tb = {(0,): "0", (-1,): "+1j * t"}
        builder = build_discretized(tb, 'x', grid=1)
        lat = next(iter(builder.sites()))[0]
        assert -1j * p['t'] == builder[lat(0), lat(1)](None, None, **p)
        assert +1j * p['t'] == builder[lat(1), lat(0)](None, None, **p)


@pytest.mark.parametrize('e_to_subs, e, subs', [
    ('A * k_x + A', '(A + B) * k_x + A + B', {'A': 'A + B'}),
])
def test_numeric_functions_with_subs(e_to_subs, e, subs):
    p = {'A': 1, 'B': 2}
    builder_direct = discretize(e)
    builder_subs = discretize(e_to_subs, locals=subs)

    lat = next(iter(builder_direct.sites()))[0]
    assert builder_direct[lat(0)](None, **p) == builder_subs[lat(0)](None, **p)

    hop_direct = builder_direct[lat(0), lat(1)](None, None, **p)
    hop_subs = builder_subs[lat(0), lat(1)](None, None, **p)
    assert hop_direct == hop_subs


def test_onsite_hopping_function_name():
    template = str(discretize('A * k_x'))
    assert 'def hopping' in template


def test_numeric_functions_advance():
    hams = [
        kx**2,
        kx**2 + x,
        fA(x),
        kx*fA(x)*kx,
        sympy.Matrix([[kx * fA(x) * kx, fA(x)*kx], [kx*fA(x), fA(x)+B]]),
        kx**2 + B * x,
        'k_x**2 + sin(x)',
        B ** 0.5 * kx**2,
        B ** (1/2) * kx**2,
        sympy.sqrt(B) * kx**2,

    ]
    for hamiltonian in hams:
        for a in [1, 2, 5]:
            for func in [lambda x: x, lambda x: x**2, lambda x: x**3]:
                symbolic, coords = discretize_symbolic(hamiltonian, 'x')
                builder = build_discretized(symbolic, coords, grid=a)
                lat = next(iter(builder.sites()))[0]

                p = dict(A=func, B=5, sin=np.sin)

                # test onsite
                v = symbolic.pop((0,)).subs({sympy.symbols('a_x'): a, B: p['B']})
                f_sym = sympy.lambdify(['A', 'x'], v)
                f_num = builder[lat(0)]

                if callable(f_num):
                    f_num = swallows_extra_kwargs(f_num)
                    for n in range(-100, 100, 10):
                        assert np.allclose(f_sym(func, a*n), f_num(lat(n), **p))
                else:
                    for n in range(-100, 100, 10):
                        assert np.allclose(f_sym(func, a*n), f_num)


                # test hoppings
                for k, v in symbolic.items():
                    v = v.subs({sympy.symbols('a_x'): a, B: p['B']})
                    f_sym = sympy.lambdify(['A', 'x'], v)
                    f_num = builder[lat(0), lat(k[0])]

                    if callable(f_num):
                        f_num = swallows_extra_kwargs(f_num)
                        for n in range(10):
                            lhs = f_sym(func, a * n)
                            rhs = f_num(lat(n), lat(n+k[0]), **p)
                            assert np.allclose(lhs, rhs)
                    else:
                        for n in range(10):
                            lhs = f_sym(fA, a * n)
                            rhs = f_num
                            assert np.allclose(lhs, rhs)


def test_numeric_functions_with_parameter():

    hamiltonian = kx**2 + fA(B, x)

    for a in [1, 2, 5]:
        for func in [lambda c, x: x+c, lambda c, x: x**2 + c]:
            symbolic, coords = discretize_symbolic(hamiltonian, 'x')
            builder = build_discretized(symbolic, coords, grid=a)
            lat = next(iter(builder.sites()))[0]

            p = dict(A=func, B=5)

            # test onsite
            v = symbolic.pop((0,)).subs({sympy.symbols('a_x'): a, B: p['B']})
            f_sym = sympy.lambdify(['A', 'x'], v)

            f_num = builder[lat(0)]
            if callable(f_num):
                f_num = swallows_extra_kwargs(f_num)

            for n in range(10):
                s = lat(n)
                xi = a * n
                if callable(f_num):
                    assert np.allclose(f_sym(func, xi), f_num(s, **p))
                else:
                    assert np.allclose(f_sym(func, xi), f_num)

            # test hoppings
            for k, v in symbolic.items():
                v = v.subs({sympy.symbols('a_x'): a, B: p['B']})
                f_sym = sympy.lambdify(['A', 'x'], v)
                f_num = builder[lat(0), lat(k[0])]

                if callable(f_num):
                    f_num = swallows_extra_kwargs(f_num)

                for n in range(10):
                    s = lat(n)
                    xi = a * n

                    lhs = f_sym(func, xi)
                    if callable(f_num):
                        rhs = f_num(lat(n), lat(n+k[0]), **p)
                    else:
                        rhs = f_num

                    assert np.allclose(lhs, rhs)


###### test grid parameter
@pytest.mark.parametrize('ham, grid_spacing, grid', [
    ('k_x', None, Monatomic([[1, ]], norbs=1)),
    ('k_x * sigma_z', None, Monatomic([[1, ]], norbs=2)),
    ('k_x', 0.5, Monatomic([[0.5, ]], norbs=1)),
    ('k_x**2 + k_y**2', 2, Monatomic([[2, 0], [0, 2]], norbs=1)),
])
def test_grid(ham, grid_spacing, grid):
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        t1 = discretize(ham, grid_spacing=grid_spacing)
        t2 = discretize(ham, grid=grid_spacing)
        t3 = discretize(ham, grid=grid)
    assert t1.lattice == t2.lattice == t3.lattice


@pytest.mark.parametrize('ham, grid_offset, offset, norbs', [
    ('k_x', None, 0, None),
    ('k_x', None, 0, 1),
    ('k_x * eye(2)', None, 0, 2),
    ('k_x', (0,), 0, None),
    ('k_x', (1,), 1, None),
    ('k_x + k_y', None, (0, 0), None),
    ('k_x + k_y', (0, 0), (0, 0), None),
    ('k_x + k_y', (1, 2), (1, 2), None),
])
def test_grid_input(ham, grid_offset, offset, norbs):
    # build appriopriate grid
    if isinstance(offset, int):
        prim_vecs = [[1, ]]
    else:
        prim_vecs = np.eye(len(offset))
    grid = Monatomic(prim_vecs, offset=grid_offset, norbs=norbs)

    tmp = discretize(ham, grid=grid)
    assert np.allclose(tmp.lattice.offset, offset)
    assert tmp.lattice.norbs == norbs

    tb_ham, coords = discretize_symbolic(ham)
    tmp = build_discretized(
        tb_ham, coords, grid=grid
    )
    assert np.allclose(tmp.lattice.offset, offset)
    assert tmp.lattice.norbs == norbs


def test_grid_offset_passed_to_functions():
    V = lambda x: x
    grid = Monatomic([[1, ]], offset=[0.5, ])
    tb = discretize('V(x)', 'x', grid=grid)
    onsite = tb[tb.lattice(0)]
    bools = [np.allclose(onsite(tb.lattice(i), V), V(tb.lattice(i).pos))
             for i in [0, 1, 5]]
    assert all(bools)


@pytest.mark.parametrize("ham, coords, grid", [
    ("k_x", None, Monatomic([[1, 0]])),
    ("k_x", 'xy', Monatomic([[1, 0]])),
    ("k_x", None, Monatomic([[1, ]], norbs=2)),
    ("k_x * eye(2)", None, Monatomic([[1, ]], norbs=1)),
    ("k_x+k_y", None, Monatomic([[1, 0], [1, 1]])),
])
def test_grid_constraints(ham, coords, grid):
    with pytest.raises(ValueError):
        discretize(ham, coords, grid=grid)


@pytest.mark.parametrize('name', ['1', '1a', '-a', '+a', 'while', 'for'])
def test_check_symbol_names(name):
    with pytest.raises(ValueError):
        discretize(sympy.Symbol(name), 'x')


def test_rectangular_grid():
    lat = Monatomic([[1, 0], [0, 2]])

    tb = discretize("V(x, y)", 'xy', grid=lat)
    assert np.allclose(tb[lat(0, 0)](lat(1, 0), lambda x, y: x), 1)
    assert np.allclose(tb[lat(0, 0)](lat(0, 1), lambda x, y: y), 2)

    tb = discretize('k_x**2 + k_y**2', grid=lat)
    assert np.allclose(tb[lat(0, 0), lat(1, 0)], -1)
    assert np.allclose(tb[lat(0, 0), lat(0, 1)], -1/4)



# Copyright 2011-2017 Kwant authors.
#
# This file is part of Kwant.  It is subject to the license terms in the file
# LICENSE.rst found in the top-level directory of this distribution and at
# http://kwant-project.org/license.  A list of Kwant authors can be found in
# the file AUTHORS.rst at the top-level directory of this distribution and at
# http://kwant-project.org/authors.

import functools as ft
from collections import deque
import pickle
import numpy as np
import tinyarray as ta
import numpy.linalg as la
from scipy.sparse import coo_matrix
import pytest
from pytest import raises
# needed to get round odd bug in test_mask_interpolate
from contextlib import contextmanager

import kwant
from kwant import operator as ops


sigma0 = np.array([[1, 0], [0, 1]])
sigmax = np.array([[0, 1], [1, 0]])
sigmay = np.array([[0, -1j], [1j, 0]])
sigmaz = np.array([[1, 0], [0, -1]])

opservables = (ops.Density, ops.Current, ops.Source)


def _random_square_system(N, norbs=1):

    def random_onsite(i):
        return 2 + kwant.digest.uniform(i.tag)

    def random_hopping(i, j):
        return -1 + kwant.digest.uniform(i.tag + j.tag)

    lat = kwant.lattice.square(norbs=norbs)
    syst = kwant.Builder()
    syst[(lat(i, j) for i in range(N) for j in range(N))] = random_onsite
    syst[lat.neighbors()] = random_hopping
    return lat, syst


def _perfect_lead(N, norbs=1):
    lat = kwant.lattice.square(norbs=norbs)
    syst = kwant.Builder(kwant.TranslationalSymmetry((-1, 0)))
    syst[(lat(0, j) for j in range(N))] = 2
    syst[lat.neighbors()] = -1
    return  lat, syst


def test_operator_construction():
    lat, syst = _random_square_system(3)
    fsyst = syst.finalized()
    N = len(fsyst.sites)

    # test construction failure if norbs not given
    latnone = kwant.lattice.chain()
    syst[latnone(0)] = 1
    for A in opservables:
        raises(ValueError, A, syst.finalized())
    del syst[latnone(0)]

    # test construction failure when dimensions of onsite do not match
    for A in opservables:
        raises(ValueError, A, fsyst, onsite=np.eye(2))

    # test that error is raised when input array is the wrong size
    for A in opservables:
        a = A(fsyst)
        kets = list(map(np.ones, [(0,), (N - 1,), (N + 1,), (N, 1)]))
        for ket in kets:
            raises(ValueError, a, ket)
            raises(ValueError, a, ket, ket)
            raises(ValueError, a.act, ket)

    # Test failure on non-hermitian
    for A in (ops.Density, ops.Current, ops.Source):
        raises(ValueError, A, fsyst, 1j)

    # Test output dtype
    ket = np.ones(len(fsyst.sites))
    for A in (ops.Density, ops.Current, ops.Source):
        a = A(fsyst)
        a_nonherm = A(fsyst, check_hermiticity=False)
        assert a(ket, ket).dtype == np.complex128
        assert a(ket).dtype == np.float64
        assert a_nonherm(ket, ket).dtype == np.complex128
        assert a_nonherm(ket).dtype == np.complex128

    # test construction with different numbers of orbitals
    lat2 = kwant.lattice.chain(norbs=2)
    extra_sites = [lat2(i) for i in range(N)]
    syst[extra_sites] = np.eye(2)
    syst[zip(fsyst.sites, extra_sites)] = ta.matrix([1, 1])
    for A in opservables:
        raises(ValueError, A, syst.finalized(), onsite=np.eye(2))
        A(syst.finalized())
        A.onsite == np.eye(2)
    del syst[extra_sites]

    check = [(ops.Density, np.arange(N).reshape(-1, 1), ta.identity(1)),
             (ops.Current, np.array(list(fsyst.graph)), ta.identity(1)),
             (ops.Source, np.arange(N).reshape(-1, 1), ta.identity(1))]

    # test basic construction
    for A, where, onsite in check:
        a = A(fsyst)
        assert np.all(np.asarray(a.where) == where)
        assert all(a.onsite == onsite for i in range(N))

    # test construction with dict `onsite`
    for A in opservables:
        B = A(fsyst, {lat: 1})
        assert all(B.onsite(i) == 1 for i in range(N))

    # test construction with a functional onsite
    for A in opservables:
        B = A(fsyst, lambda site: site.pos[0])  # x-position operator
        assert all(B.onsite(i) == fsyst.sites[i].pos[0] for i in range(N))

    # test construction with `where` given by a sequence
    where = [lat(2, 2), lat(1, 1)]
    fwhere = tuple(fsyst.id_by_site[s] for s in where)
    A = ops.Density(fsyst, where=where)
    assert np.all(np.asarray(A.where).reshape(-1) == fwhere)

    where = [(lat(2, 2), lat(1, 2)), (lat(0, 0), lat(0, 1))]
    fwhere = np.asarray([(fsyst.id_by_site[a], fsyst.id_by_site[b])
              for a, b in where])
    A = ops.Current(fsyst, where=where)
    assert np.all(np.asarray(A.where) == fwhere)

    # test construction with `where` given by a function
    tag_list = [(1, 0), (1, 1), (1, 2)]
    def where(site):
        return site.tag in tag_list
    A = ops.Density(fsyst, where=where)
    assert all(fsyst.sites[A.where[w, 0]].tag in tag_list
               for w in range(A.where.shape[0]))

    where_list = set(kwant.HoppingKind((1, 0), lat)(syst))
    fwhere_list = set((fsyst.id_by_site[a], fsyst.id_by_site[b])
                      for a, b in where_list)
    def where(a, b):
        return (a, b) in where_list
    A = ops.Current(fsyst, where=where)
    assert all((a, b) in fwhere_list for a, b in A.where)

    # test that `sum` is passed to constructors correctly
    for A in opservables:
        A(fsyst, sum=True).sum == True


def _test(A, bra, ket=None, per_el_val=None, reduced_val=None, params=None):
    if per_el_val is not None:
        val = A(bra, ket, params=params)
        assert np.allclose(val, per_el_val)
        # with bound args
        val = A.bind(params=params)(bra, ket)
        assert np.allclose(val, per_el_val)
    # test that inner products give the same thing
    ket = bra if ket is None else ket
    act_val = np.dot(bra.conj(), A.act(ket, params=params))
    inner_val = np.sum(A(bra, ket, params=params))
    # check also when sum is done internally by operator
    try:
        sum_reset = A.sum
        A.sum = True
        sum_inner_val = A(bra, ket, params=params)
        assert inner_val == sum_inner_val
    finally:
        A.sum = sum_reset

    assert np.isclose(act_val, inner_val)
    assert np.isclose(sum_inner_val, inner_val)
    if reduced_val is not None:
        assert np.isclose(inner_val, reduced_val)


def test_opservables_finite():
    lat, syst = _random_square_system(3)
    fsyst = syst.finalized()
    ev, wfs = la.eigh(fsyst.hamiltonian_submatrix())

    Q = ops.Density(fsyst)
    Qtot = ops.Density(fsyst, sum=True)
    J = ops.Current(fsyst)
    K = ops.Source(fsyst)

    for i, wf in enumerate(wfs.T):  # wfs[:, i] is i'th eigenvector
        assert np.allclose(Q.act(wf), wf)  # this operation is identity
        _test(Q, wf, reduced_val=1)  # eigenvectors are normalized
        _test(Qtot, wf, per_el_val=1)  # eigenvectors are normalized
        _test(J, wf, per_el_val=0)  # time-reversal symmetry: no current
        _test(K, wf, per_el_val=0)  # onsite commutes with hamiltonian

    # check that we get correct (complex) output
    for bra, ket in zip(wfs.T, wfs.T):
        _test(Q, bra, ket, per_el_val=(bra * ket))

    # check with get_hermiticity=False
    Qi = ops.Density(fsyst, 1j, check_hermiticity=False)
    for wf in wfs.T:  # wfs[:, i] is i'th eigenvector
        assert np.allclose(Qi.act(wf), 1j * wf)

    # test with different numbers of orbitals
    lat2 = kwant.lattice.chain(norbs=2)
    extra_sites = [lat2(i) for i in range(len(fsyst.sites))]
    syst[extra_sites] = np.eye(2)
    syst[zip(fsyst.sites, extra_sites)] = ta.matrix([1, 1])
    fsyst = syst.finalized()
    ev, wfs = la.eigh(fsyst.hamiltonian_submatrix())

    Q = ops.Density(fsyst)
    Qtot = ops.Density(fsyst, sum=True)
    J = ops.Current(fsyst)
    K = ops.Source(fsyst)

    for wf in wfs.T:  # wfs[:, i] is i'th eigenvector
        assert np.allclose(Q.act(wf), wf)  # this operation is identity
        _test(Q, wf, reduced_val=1)  # eigenvectors are normalized
        _test(Qtot, wf, per_el_val=1)  # eigenvectors are normalized
        _test(J, wf, per_el_val=0)  # time-reversal symmetry: no current
        _test(K, wf, per_el_val=0)  # onsite commutes with hamiltonian


def test_opservables_infinite():
    lat, lead = _perfect_lead(3)
    flead = lead.finalized()

    # Cannot calculate current between unit cells because the wavefunction
    # is only defined over a single unit cell
    raises(ValueError, ops.Current, flead, where=[(lat(0, j), lat(1, j))
                                                   for j in range(3)])

    transverse_kind = kwant.builder.HoppingKind((0, 1), lat)
    J_intra = ops.Current(flead, where=list(transverse_kind(lead)))

    prop, _ = flead.modes(energy=1.)
    for wf, v in zip(prop.wave_functions.T, prop.velocities):
        _test(J_intra, wf, per_el_val=0)  # no transverse current


def test_opservables_scattering():
    # Disordered system with two ordered strips on the left/right.  We check
    # that the current on the right of the disorder due to incoming mode `m` is
    # equal to Σ_n |t_nm|^2. Similarly the current on the left of the disorder
    # is checked against 1 - Σ_n |r_nm|^2

    N = 10
    lat, syst = _random_square_system(N)
    # add extra sites so we can calculate the current in a region
    # where there is no backscattering
    syst[(lat(i, j) for i in [-1, N] for j in range(N))] = 2
    syst[((lat(-1, j), lat(0, j)) for j in range(N))] = -1
    syst[((lat(N-1, j), lat(N, j)) for j in range(N))] = -1

    lat, lead = _perfect_lead(3)
    syst.attach_lead(lead)
    syst.attach_lead(lead.reversed())
    fsyst = syst.finalized()

    # currents on the left and right of the disordered region
    right_interface = [(lat(N, j), lat(N-1, j)) for j in range(3)]
    left_interface = [(lat(0, j), lat(-1, j)) for j in range(3)]
    J_right = ops.Current(fsyst, where=right_interface)
    J_right_tot = ops.Current(fsyst, where=right_interface, sum=True)
    J_left = ops.Current(fsyst, where=left_interface)
    J_left_tot = ops.Current(fsyst, where=left_interface, sum=True)

    smatrix = kwant.smatrix(fsyst, energy=1.0)
    t = smatrix.submatrix(1, 0).T  # want to iterate over the columns
    r = smatrix.submatrix(0, 0).T  # want to iterate over the columns
    wfs = kwant.wave_function(fsyst, energy=1.0)(0)
    for rv, tv, wf in zip(r, t, wfs):
        _test(J_right, wf, reduced_val=np.sum(np.abs(tv)**2))
        _test(J_right_tot, wf, per_el_val=np.sum(np.abs(tv)**2))
        _test(J_left, wf, reduced_val=(1 - np.sum(np.abs(rv)**2)))
        _test(J_left_tot, wf, per_el_val=(1 - np.sum(np.abs(rv)**2)))


def test_opservables_spin():

    def onsite(site, B):
        return 2 * np.eye(2) + B * sigmaz

    L = 20
    lat = kwant.lattice.chain(norbs=2)
    syst = kwant.Builder()
    syst[(lat(i) for i in range(L))] = onsite
    syst[lat.neighbors()] = -1 * np.eye(2)
    lead = kwant.Builder(kwant.TranslationalSymmetry((-1,)))
    lead[lat(0)] = onsite
    lead[lat.neighbors()] = -1 * np.eye(2)
    syst.attach_lead(lead)
    syst.attach_lead(lead.reversed())
    fsyst = syst.finalized()
    params = dict(B=0.1)
    down, up = kwant.wave_function(fsyst, energy=1., params=params)(0)

    x_hoppings = kwant.builder.HoppingKind((1,), lat)
    spin_current_z = ops.Current(fsyst, sigmaz, where=x_hoppings(syst))
    _test(spin_current_z, up, params=params, per_el_val=1)
    _test(spin_current_z, down, params=params, per_el_val=-1)

    # calculate spin_x torque
    spin_torque_x = ops.Source(fsyst, sigmax, where=[lat(L//2)])
    i = fsyst.id_by_site[lat(L//2)]
    psi = up[2*i:2*(i+1)] + down[2*i:2*(i+1)]
    H_ii = onsite(None, **params)
    K = np.dot(H_ii, sigmax) - np.dot(sigmax, H_ii)
    expect = 1j * ft.reduce(np.dot, (psi.conj(), K, psi))
    _test(spin_torque_x, up+down, params=params, reduced_val=expect)


def test_opservables_gauged():
    # Test that we get the same answer when we apply a random
    # gauge (unitary) transformation to each site. We also
    # adjust our definition of the current to match

    # this is to get round a bug in test_mask_interpolate (?!) that fails when
    # the random number state is altered.
    @contextmanager
    def save_random_state():
        old_state = np.random.get_state()
        yield
        np.random.set_state(old_state)

    L = 20
    with save_random_state():
        Us = deque([kwant.rmt.circular(2) for i in range(L)])
    # need these to get the coupling to the leads right
    Us.append(np.eye(2))
    Us.appendleft(np.eye(2))

    H0 = 2 * np.eye(2) + 0.1 * sigmaz  # onsite
    V0 = -1 * np.eye(2)  # hopping

    lat = kwant.lattice.chain(norbs=2)
    syst = kwant.Builder()

    for i, U in enumerate(Us):
        syst[lat(i)] = ft.reduce(np.dot, (U, H0, U.conjugate().transpose()))

    for a, b in kwant.builder.HoppingKind((1,), lat)(syst):
        i, j = a.tag[0], b.tag[0]
        syst[(a, b)] = ft.reduce(np.dot,
                                 (Us[i], V0, Us[j].conjugate().transpose()))

    lead = kwant.Builder(kwant.TranslationalSymmetry((-1,)))
    lead[lat(0)] = H0
    lead[lat.neighbors()] = V0
    syst.attach_lead(lead)
    syst.attach_lead(lead.reversed())
    fsyst = syst.finalized()
    down, up = kwant.wave_function(fsyst, energy=1.0)(0)

    def M_a(site):
        i = site.tag[0]
        return ft.reduce(np.dot,
                         (Us[i], sigmaz, Us[i].conjugate().transpose()))

    x_hoppings = kwant.builder.HoppingKind((1,), lat)
    spin_current_gauge = ops.Current(fsyst, M_a, where=x_hoppings(syst))
    _test(spin_current_gauge, up, per_el_val=1)
    _test(spin_current_gauge, down, per_el_val=-1)
    # check the reverse is also true
    minus_x_hoppings = kwant.builder.HoppingKind((-1,), lat)
    spin_current_gauge = ops.Current(fsyst, M_a, where=minus_x_hoppings(syst))
    _test(spin_current_gauge, up, per_el_val=-1)
    _test(spin_current_gauge, down, per_el_val=1)


def test_tocoo():
    syst = kwant.Builder()
    lat1 = kwant.lattice.chain(norbs=1)
    syst[lat1(0)] = syst[lat1(1)] = 0
    syst = syst.finalized()

    op = ops.Density(syst)
    assert isinstance(op.tocoo(), coo_matrix)

    # Constant and non-constant values.
    assert np.all(op.tocoo().toarray() == np.eye(2))
    op = ops.Density(syst, lambda site: 1)
    assert np.all(op.tocoo().toarray() == np.eye(2))

    # Correct treatment of where
    op = ops.Density(syst, where=[lat1(0)])
    assert np.all(op.tocoo().toarray() == [[1, 0], [0, 0]])

    # No accidental transpose.
    syst = kwant.Builder()
    lat2 = kwant.lattice.chain(norbs=2)
    syst[lat2(0)] = lambda site, p: np.eye(2)
    syst = syst.finalized()
    op = ops.Density(syst, [[1, 1], [0, 1]], check_hermiticity=False)
    assert np.all(op.tocoo().toarray() == [[1, 1], [0, 1]])

    op = ops.Density(syst, lambda site, p: [[1, 1], [0, 1]],
                     check_hermiticity=False)
    op = op.bind(params=dict(p=1))
    raises(ValueError, op.tocoo, [1])


# We need to keep testing 'args', but we don't want to see
# all the deprecation warnings in the test logs
@pytest.mark.filterwarnings("ignore:.*'args' parameter")
@pytest.mark.parametrize("A", opservables)
def test_arg_passing(A):
    lat1 = kwant.lattice.chain(norbs=1)

    syst = kwant.Builder()
    syst[lat1(0)] = syst[lat1(1)] = lambda s0, a, b: s0.pos + a + b
    syst[lat1.neighbors()] = lambda s0, s1, a, b: a - b
    fsyst = syst.finalized()

    wf = np.ones(len(fsyst.sites))

    # test missing params
    op = A(fsyst, onsite=lambda x, a, b: 1)
    params = dict(a=1)
    with raises(TypeError) as exc:
        op(wf, params=params)
    with raises(TypeError) as exc:
        op.act(wf, params=params)
    with raises(TypeError) as exc:
        op.bind(params=params)
    if hasattr(op, 'tocoo'):
        with raises(TypeError) as exc:
            op.tocoo(params=params)


    op = A(fsyst)
    canonical_args = (1, 2)
    params = dict(a=1, b=2)
    call_should_be = op(wf, args=canonical_args)
    act_should_be = op.act(wf, args=canonical_args)
    has_tocoo = hasattr(op, 'tocoo')
    if has_tocoo:
        tocoo_should_be = op.tocoo(args=canonical_args).toarray()

    with raises(TypeError) as exc:
        op(wf, args=canonical_args, params=params)
    assert 'mutually exclusive' in str(exc)
    with raises(TypeError) as exc:
        op.act(wf, args=canonical_args, params=params)
    assert 'mutually exclusive' in str(exc)
    with raises(TypeError) as exc:
        op.bind(args=canonical_args, params=params)
    assert 'mutually exclusive' in str(exc)
    if has_tocoo:
        with raises(TypeError) as exc:
            op.tocoo(args=canonical_args, params=params)
        assert 'mutually exclusive' in str(exc)

    np.testing.assert_array_equal(
        call_should_be, op(wf, params=params))
    np.testing.assert_array_equal(
        act_should_be, op.act(wf, params=params))
    if has_tocoo:
        np.testing.assert_array_equal(
            tocoo_should_be, op.tocoo(params=params).toarray())
    # after binding
    op2 = op.bind(params=params)
    np.testing.assert_array_equal(
        call_should_be, op2(wf))
    np.testing.assert_array_equal(
        act_should_be, op2.act(wf))
    if has_tocoo:
        np.testing.assert_array_equal(
            tocoo_should_be, op2.tocoo().toarray())

    # system and onsite having different args
    def onsite(site, flip):
        return -1 if flip else 1

    op = A(fsyst, onsite=onsite)
    params['flip'] = True
    call_should_be = -call_should_be
    act_should_be = -act_should_be
    if has_tocoo:
        tocoo_should_be = -tocoo_should_be

    np.testing.assert_array_equal(
        call_should_be, op(wf, params=params))
    np.testing.assert_array_equal(
        act_should_be, op.act(wf, params=params))
    if has_tocoo:
        np.testing.assert_array_equal(
            tocoo_should_be, op.tocoo(params=params).toarray())
    # after binding
    op2 = op.bind(params=params)
    np.testing.assert_array_equal(
        call_should_be, op2(wf))
    np.testing.assert_array_equal(
        act_should_be, op2.act(wf))
    if has_tocoo:
        np.testing.assert_array_equal(
            tocoo_should_be, op2.tocoo().toarray())


def random_onsite(i):
    return (2 + kwant.digest.uniform(i.tag)) * sigmaz


def random_hopping(i, j):
    return (-1 + kwant.digest.uniform(i.tag + j.tag)) * sigmay


def f_sigmay(i):
    return sigma0


@pytest.mark.parametrize("A", opservables)
def test_pickling(A):

    lat = kwant.lattice.square(norbs=2)
    syst = kwant.Builder()
    syst[(lat(i, j) for i in range(5) for j in range(5))] = random_onsite
    syst[lat.neighbors()] = random_hopping
    fsyst = syst.finalized()

    wf = np.random.rand(2 * len(fsyst.sites))

    def all_equal(it):
        it = iter(it)
        first = next(it)
        return all(np.all(first == rest) for rest in it)

    ops = [
        A(fsyst),
        A(fsyst, onsite=f_sigmay),
        A(fsyst, onsite=sigmaz),
        A(fsyst, sum=True),
        A(fsyst, onsite=sigmaz, sum=True),
        A(fsyst, onsite=f_sigmay, sum=True),
    ]
    ops += [op.bind() for op in ops]

    for op in ops:
        loaded_op = pickle.loads(pickle.dumps(op))
        assert np.all(op(wf) == loaded_op(wf))
